@@ -25,13 +25,13 @@
 # OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 
-import datetime
 import errno
 import getopt
 import ipaddress
 import os
 import signal
 import socket
+import ssl
 import sys
 import time
 from statistics import stdev
@@ -51,7 +51,8 @@ import httpx
 
 from dnsdiag.dns import PROTO_UDP, PROTO_TCP, PROTO_TLS, PROTO_HTTPS, PROTO_QUIC, PROTO_HTTP3, proto_to_text, \
     get_default_port, valid_rdatatype
-from dnsdiag.shared import __version__, valid_hostname, unsupported_feature, random_string, die, err
+from dnsdiag.shared import __version__, valid_hostname, unsupported_feature, random_string, die, err, \
+    set_protocol_exclusive
 
 __author__ = 'Babak Farrokhi (babak@farrokhi.net)'
 __license__ = 'BSD'
@@ -187,6 +188,7 @@ def main() -> None:
     af = None
     af_ipv4_set = False
     af_ipv6_set = False
+    proto_option_set: Optional[str] = None
     qname = 'wikipedia.org'
 
     try:
@@ -266,27 +268,27 @@ def main() -> None:
                 die(f"ERROR: invalid RR class: {a}")
 
         elif o in ("-T", "--tcp"):
-            proto = PROTO_TCP
+            proto, proto_option_set = set_protocol_exclusive(PROTO_TCP, o, proto_option_set)
             if use_default_dst_port:
                 dst_port = get_default_port(proto)
 
         elif o in ("-X", "--tls"):
-            proto = PROTO_TLS
+            proto, proto_option_set = set_protocol_exclusive(PROTO_TLS, o, proto_option_set)
             if use_default_dst_port:
                 dst_port = get_default_port(proto)
 
         elif o in ("-H", "--doh"):
-            proto = PROTO_HTTPS
+            proto, proto_option_set = set_protocol_exclusive(PROTO_HTTPS, o, proto_option_set)
             if use_default_dst_port:
                 dst_port = get_default_port(proto)
 
         elif o in ("-Q", "--quic"):
-            proto = PROTO_QUIC
+            proto, proto_option_set = set_protocol_exclusive(PROTO_QUIC, o, proto_option_set)
             if use_default_dst_port:
                 dst_port = get_default_port(proto)
 
         elif o in ("-3", "--http3"):
-            proto = PROTO_HTTP3
+            proto, proto_option_set = set_protocol_exclusive(PROTO_HTTP3, o, proto_option_set)
             if use_default_dst_port:
                 dst_port = get_default_port(proto)
 
@@ -427,11 +429,30 @@ def main() -> None:
                                         source=src_ip, source_port=src_port)
             elif proto is PROTO_TLS:
                 if hasattr(dns.query, 'tls'):
-                    # Use resolved IP for connection, but provide hostname for SNI/certificate validation
-                    server_hostname = dnsserver_hostname if dnsserver_hostname != dnsserver_ip else None
-                    answers = dns.query.tls(query, dnsserver_ip, timeout=timeout, port=dst_port,
-                                            source=src_ip, source_port=src_port,
-                                            server_hostname=server_hostname)
+                    try:
+                        # Use resolved IP for connection, but provide hostname for SNI/certificate validation
+                        server_hostname = dnsserver_hostname if dnsserver_hostname != dnsserver_ip else None
+                        answers = dns.query.tls(query, dnsserver_ip, timeout=timeout, port=dst_port,
+                                                source=src_ip, source_port=src_port,
+                                                server_hostname=server_hostname)
+                    except dns.exception.Timeout:
+                        if not quiet:
+                            print("Request timeout", flush=True)
+                        continue
+                    except (ConnectionResetError, BrokenPipeError):
+                        if not quiet:
+                            print("Connection closed unexpectedly", flush=True)
+                        continue
+                    except socket.gaierror:
+                        if not quiet:
+                            print("Name resolution failed", flush=True)
+                        continue
+                    except ssl.SSLCertVerificationError as e:
+                        die(f"Certificate verification failed: {e}")
+                    except ssl.SSLError:
+                        if not quiet:
+                            print("Connection failed", flush=True)
+                        continue
                 else:
                     unsupported_feature("DNS-over-TLS")
 
@@ -449,7 +470,27 @@ def main() -> None:
                     except dns.query.NoDOH:
                         die("ERROR: python httpx module not available")
                     except httpx.ConnectError:
-                        die(f"DoH connection failed on port {dst_port}")
+                        if not quiet:
+                            print("Connection refused", flush=True)
+                        continue
+                    except (ConnectionResetError, BrokenPipeError):
+                        if not quiet:
+                            print("Connection closed unexpectedly", flush=True)
+                        continue
+                    except socket.gaierror:
+                        if not quiet:
+                            print("Name resolution failed", flush=True)
+                        continue
+                    except (httpx.WriteTimeout, httpx.StreamError, httpx.ProtocolError):
+                        if not quiet:
+                            print("Connection failed", flush=True)
+                        continue
+                    except ssl.SSLCertVerificationError as e:
+                        die(f"Certificate verification failed: {e}")
+                    except ssl.SSLError:
+                        if not quiet:
+                            print("Connection failed", flush=True)
+                        continue
                 else:
                     unsupported_feature("DNS-over-HTTPS (DoH)")
 
@@ -465,7 +506,35 @@ def main() -> None:
                                                   source=src_ip, source_port=src_port,
                                                   http_version=dns.query.HTTPVersion.H3)
                     except ConnectionRefusedError:
-                        die(f"DoH3 connection refused on port {dst_port}")
+                        if not quiet:
+                            print("Connection refused", flush=True)
+                        continue
+                    except (ConnectionResetError, BrokenPipeError):
+                        if not quiet:
+                            print("Connection closed unexpectedly", flush=True)
+                        continue
+                    except socket.gaierror:
+                        if not quiet:
+                            print("Name resolution failed", flush=True)
+                        continue
+                    except (httpx.WriteTimeout, httpx.StreamError, httpx.ProtocolError):
+                        if not quiet:
+                            print("Connection failed", flush=True)
+                        continue
+                    except ssl.SSLCertVerificationError as e:
+                        die(f"Certificate verification failed: {e}")
+                    except ssl.SSLError:
+                        if not quiet:
+                            print("Connection failed", flush=True)
+                        continue
+                    except Exception as e:
+                        # Catch QUIC-specific exceptions (UnexpectedEOF, etc.)
+                        if e.__class__.__name__ in ('UnexpectedEOF', 'QuicConnectionError', 'H3Error'):
+                            if not quiet:
+                                print("Connection closed unexpectedly", flush=True)
+                            continue
+                        else:
+                            raise
                 else:
                     unsupported_feature("DNS-over-HTTPS/3 (DoH3)")
 
@@ -478,9 +547,35 @@ def main() -> None:
                                                  source=src_ip, source_port=src_port,
                                                  server_hostname=server_hostname)
                     except dns.exception.Timeout:
-                        die(f"DoQ connection timeout on port {dst_port}")
+                        if not quiet:
+                            print("Request timeout", flush=True)
+                        continue
                     except ConnectionRefusedError:
-                        die(f"DoQ connection refused on port {dst_port}")
+                        if not quiet:
+                            print("Connection refused", flush=True)
+                        continue
+                    except (ConnectionResetError, BrokenPipeError):
+                        if not quiet:
+                            print("Connection closed unexpectedly", flush=True)
+                        continue
+                    except socket.gaierror:
+                        if not quiet:
+                            print("Name resolution failed", flush=True)
+                        continue
+                    except ssl.SSLCertVerificationError as e:
+                        die(f"Certificate verification failed: {e}")
+                    except ssl.SSLError:
+                        if not quiet:
+                            print("Connection failed", flush=True)
+                        continue
+                    except Exception as e:
+                        # Catch QUIC-specific exceptions
+                        if e.__class__.__name__ in ('UnexpectedEOF', 'QuicConnectionError', 'StreamFinishedError'):
+                            if not quiet:
+                                print("Connection closed unexpectedly", flush=True)
+                            continue
+                        else:
+                            raise
                 else:
                     unsupported_feature("DNS-over-QUIC (DoQ)")
 
@@ -525,13 +620,8 @@ def main() -> None:
             shutdown = True
             break
         else:
-            # convert time to milliseconds, considering that
-            # time property is returned differently by query.https
-            # dns library returns float for most protocols but timedelta for HTTPS
-            if isinstance(answers.time, datetime.timedelta):
-                elapsed = answers.time.total_seconds() * 1000
-            else:
-                elapsed = answers.time * 1000
+            # Use perf_counter() measurements for accurate wall-clock time
+            elapsed = (etime - stime) * 1000  # Convert seconds to milliseconds
             response_time.append(elapsed)
             if not quiet:
                 extras = ""
